/*
    The IDA Pro Book - SEH Chain plugin
    Copyright (C) 2008 Chris Eagle <cseagle@gmail.com>

    This program is free software; you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by the Free
    Software Foundation; either version 2 of the License, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
    more details.

    You should have received a copy of the GNU General Public License along with
    this program; if not, write to the Free Software Foundation, Inc., 59 Temple
    Place, Suite 330, Boston, MA 02111-1307 USA

*/

/*
 *  This is the SEH chain plugin.  When the debugger is active for 
 *  a Windows, x86 process, this plugin will display the Structured
 *  Exception Handling chain for the active thread.
 */

#include <ida.hpp>
#include <idp.hpp>
#include <bytes.hpp>
#include <name.hpp>
#include <loader.hpp>
#include <kernwin.hpp>
#include <segment.hpp>
#include <dbg.hpp>
#include "sdk_versions.h"

//The column headers for the SEH Chain display
char *headers[] = {
   "Address", "SEH Handler"
};

//The column widths for the SEH chain display
int widths[] = {16, 64};

ulong get_tib_start(thread_id_t tid) {
   char segname[64];
   qsnprintf(segname, sizeof(segname), "TIB[%08X]", tid);
   segment_t *seg = get_segm_by_name(segname);
   return seg ? seg->startEA : BADADDR;
}

//A function to compute the number of rows required for
//an SEH chain display
ulong idaapi seh_sizer(void *obj) {
   int count = 0;
   thread_id_t tid = (thread_id_t)obj;
   ulong start = get_tib_start(tid);
   if (start != BADADDR) {
      for (start = get_long(start); 
           start != 0xFFFFFFFF; 
           start = get_long(start)) {
         count++;
      }
   }
   return count;
}

void idaapi seh_getl(void *obj, ulong n, char* const*arrptr) {
   if (n == 0) {
      qstrncpy(arrptr[0], headers[0], strlen(headers[0]) + 1);
      qstrncpy(arrptr[1], headers[1], strlen(headers[1]) + 1);
   }
   else {
      thread_id_t tid = (thread_id_t)obj;
      ulong start = get_tib_start(tid);
      if (start != BADADDR) {
         char name[128];
         start = get_long(start);
         //Walk the SEH Chain to count the number of entries
         while (--n && start != 0xFFFFFFFF) {
            start = get_long(start);
         }
         if (start == 0xFFFFFFFF) return;
         ulong func = get_long(start + 4);
         //The address of the record is always hex
         qsnprintf(arrptr[0], 16, "%08X", start);
         //we prefer to display the handler address as a name
         char *nm = get_true_name(BADADDR, func, name, sizeof(name));
         if (nm) {
            //the handler address was already named
            qstrncpy(arrptr[1], name, sizeof(name));
         }
         else {
            //the handler address was not named, lets try to create a
            //function there (since it is an SEH function!)
            if (add_func(func, BADADDR)) {
               //Successfully created the function, it should have a name!
               get_true_name(BADADDR, func, name, sizeof(name));
               qstrncpy(arrptr[1], name, sizeof(name));
            }
            else {
               //Could not create a function, give up and print handler
               //address as a hex value.
               qsnprintf(arrptr[1], 16, "%08X", func);
            }
         }
      }
   }
}

//Function to process the enter key or double click 
//events on items in an SEH Chain display
void idaapi seh_enter(void *obj, ulong n) {
   thread_id_t tid = (thread_id_t)obj;
   ulong start = get_tib_start(tid);
   if (start != BADADDR) {
      start = get_long(start);
      //Walk the SEH Chain to count the number of entries
      while (--n && start != 0xFFFFFFFF) {
         start = get_long(start);
      }
      if (start != 0xFFFFFFFF) {
         jumpto(get_long(start + 4));
      }
   }
}

void show_thread_seh_handlers(thread_id_t tid) {
   char segname[64];
   qsnprintf(segname, sizeof(segname), "TIB[%08X]", tid);
   segment_t *seg = get_segm_by_name(segname);
   if (seg == NULL) {
      warning("%s segment not found", segname);
      return;
   }
   char title[64];
   qsnprintf(title, sizeof(title), "SEH Chain of thread %08X", tid);
   //open as non-modal info dialog
   choose2(0, -1, -1, -1, -1, (void*)tid, 2, widths,
          seh_sizer, seh_getl, title, -1, 1, NULL,
          NULL, NULL, NULL, seh_enter, NULL, NULL, NULL);
}

//--------------------------------------------------------------------------
//
//      The plugin method
//
//      This is the main function of plugin.
//
//      It will be called when the user activates the plugin.
//
//              arg - the input argument, it can be specified in
//                    plugins.cfg file. The default is zero.
//
//
void idaapi run(int arg) {
   if (dbg == NULL) {
      warning("Debugger is not active");
      return;
   }
   int threads = get_thread_qty();
   for (int i = 0; i < threads; i++) {
      show_thread_seh_handlers(getn_thread(i));
   }
}

//--------------------------------------------------------------------------
//
//      Initialize.
//
//      IDA will call this function only once.
//      If this function returns PLGUIN_SKIP, IDA will never load it again.
//      If this function returns PLUGIN_OK, IDA will unload the plugin but
//      remember that the plugin agreed to work with the database.
//      The plugin will be loaded again if the user invokes it by
//      pressing the hotkey or selecting it from the menu.
//      After the second load the plugin will stay on memory.
//      If this function returns PLUGIN_KEEP, IDA will keep the plugin
//      in the memory. In this case the initialization function can hook
//      into the processor module and user interface notification points.
//      See the hook_to_notification_point() function.
//
//      In this example we check the input file format and make the decision.
//      You may or may not check any other conditions to decide what you do:
//      whether you agree to work with the database or not.
//
int idaapi init(void) {
   if (strcmp(inf.procName, "metapc")) return PLUGIN_SKIP;
   if (inf.filetype != f_PE) return PLUGIN_SKIP;
   return PLUGIN_KEEP;
}

//--------------------------------------------------------------------------
//      Terminate.
//      Usually this callback is empty.
//      The plugin should unhook from the notification lists if
//      hook_to_notification_point() was used.
//
//      IDA will call this function when the user asks to exit.
//      This function won't be called in the case of emergency exits.

void idaapi term(void) {
   //nothing to do for our simple plugin
}

//--------------------------------------------------------------------------
char comment[] = "This plugin displays the current SEH chain.";

char help[] = "Display the current SEH chain\n";

//--------------------------------------------------------------------------
// This is the preferred name of the plugin module in the menu system
// The preferred name may be overriden in plugins.cfg file

char wanted_name[] = "SEH chain";


// This is the preferred hotkey for the plugin module
// The preferred hotkey may be overriden in plugins.cfg file
// Note: IDA won't tell you if the hotkey is not correct
//       It will just disable the hotkey.

char wanted_hotkey[] = "Alt-F5";


//--------------------------------------------------------------------------
//
//      PLUGIN DESCRIPTION BLOCK
//
//--------------------------------------------------------------------------
plugin_t PLUGIN =
{
  IDP_INTERFACE_VERSION,
  0,                    // plugin flags
  init,                 // initialize

  term,                 // terminate. this pointer may be NULL.

  run,                  // invoke plugin

  comment,              // long comment about the plugin
                        // it could appear in the status line
                        // or as a hint

  help,                 // multiline help about the plugin

  wanted_name,          // the preferred short name of the plugin
  wanted_hotkey         // the preferred hotkey to run the plugin
};
